/**************************************************************
efx_NSContrast.cpp

Copyright John Paul Chacha. All Rights Reserved.

This source code file, which has been provided by one
John Paul Chacha as part of his software product for use only
by licensed users of that product, includes proprietary
information of John Paul Chacha.

USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS
OF THE LICENSE AGREEMENT FURNISHED WITH THE PRODUCT.

IN PARTICULAR, JOHN PAUL CHACHA SHALL BE FREE FROM ANY CLAIMS
OR LIABILITIES ARISING OUT OF THE USE OR MISUSE OF THIS FILE
AND/OR ITS CONTENTS.
**************************************************************/

#include "../common/system.h"
#include "resource.h"
#include "../common/plugin.h"
#include "../common/suite_info.h"
#include "../common/win_util.h"


//=============================================================
#define PLUGIN_NAME L"Non-saturating Contrast"
#define PLUGIN_VERSION (((SUITE_VER_HI)<<16)|(SUITE_VER_LO))

#define MY_STORE_UUID (*((LPDWORD)("_NSC")))
#define MY_STORE_NAME L"efx_NSContrast.dll"

#define APP_MSG_NOTIFY						WM_APP+ 1
#define TIMER_PREVIEW_UPDATE				3897

//-------------------------------------------------------------
struct tmp_settings_tag
{
	unsigned long uuid;
	int power;
};

//-------------------------------------------------------------
HINSTANCE	api_inst;
tmp_settings_tag	tmp_settings={0,};
efx_IMAGE_T*		lp_host;

LRESULT CALLBACK msg_dlg_effect(
	HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar);

//-------------------------------------------------------------


//=============================================================
//=============================================================



/**************************************************************
This function is called during dll startup/shutdown.
**************************************************************/
BOOL APIENTRY DllMain(HANDLE hModule,
	unsigned long ul_reason_for_call,LPVOID lpReserved)
{
	switch(ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		api_inst=(HINSTANCE)hModule;
		break;
	case DLL_THREAD_ATTACH:
		break;
	case DLL_THREAD_DETACH:
		break;
	case DLL_PROCESS_DETACH:
		break;
	}
	UNREFERENCED_PARAMETER(lpReserved);
	return TRUE;
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


/**************************************************************
This function processes messages for the about dialog
**************************************************************/
LRESULT CALLBACK msg_dlg_about(
	HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		SetWindowText(hDlg,L"About " PLUGIN_NAME);
		SetDlgItemText(hDlg,IDC_STATIC1,
			PLUGIN_NAME L" (" SUITE_LOGODATE L")\n\n"

			L"Created by John Paul Chacha's Lab. "
			L"Tested with "SUITE_FULLNAME L".\n\n"

			L"No warranties are made or implied."
			);
		window_Center(hDlg,GetParent(hDlg));
		return 1|lpar;
	case WM_COMMAND:
		switch(LOWORD(wpar))
		{
		case IDOK    :
			EndDialog(hDlg, 1);
			break;
		case IDCANCEL:
			EndDialog(hDlg,-1);
			break;
		}
		return 1;
	}
	return 0;
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


/**************************************************************
This function is called to display plugin's about dialog box
**************************************************************/
int plg_About(HWND parent)
{
	DialogBoxParam(api_inst,(LPCTSTR)IDD_ABOUT,parent,
		(DLGPROC)msg_dlg_about,0);
	return 1;
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


/**************************************************************
This function is called to display plugin's config dialog box
**************************************************************/
int plg_Configure(HWND parent)
{
	return -1;
	UNREFERENCED_PARAMETER(parent);
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


/**************************************************************
This function is called to query plugin's name and type.
**************************************************************/
int plg_GetInfo(plg_INFO* info)
{
	info->api_type=PLG_TYPE_IMAGE_EFFECT;
	info->api_version=PLUGIN_INTERFACE_VERSION;
	info->plugin_version=PLUGIN_VERSION;
	wcsncpy(info->plugin_name,PLUGIN_NAME,63);
	info->plugin_name[63]=0;
	return 1;
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


/**************************************************************
This is the dll function that Chasys Draw IES calls to perform
the effect exposed by this plugin.
**************************************************************/
int efx_DoEffect(efx_IMAGE_T* data)
{
	int i,ret;

	//---------------------------------------------------------
	if(!data->pi_BasicImage) return 0;
	if(data->pi_BasicImage->version<PI_BASICIMAGE_VERSION)
		return 0;

	//---------------------------------------------------------
	tmp_settings.power=100;
	if(data->pi_StateStore)
	{
		i=data->pi_StateStore->size(MY_STORE_NAME);
		if(i>=sizeof(tmp_settings_tag))
		{
			data->pi_StateStore->read(MY_STORE_NAME,
				&tmp_settings,sizeof(tmp_settings_tag));
		}
	}

	//---------------------------------------------------------
	lp_host=data;
	lp_host->load();
	ret=DialogBoxParam(api_inst,(LPCTSTR)IDD_EFFECT,
		data->hwnd,(DLGPROC)msg_dlg_effect,0);

	//---------------------------------------------------------
	if(data->pi_StateStore)
	{
		data->pi_StateStore->write(MY_STORE_NAME,
			&tmp_settings,sizeof(tmp_settings_tag));
	}
	if(!ret) return -99;
	return 1;
}
//=============================================================
int effect_Execute(int power)
{
	int i;
	unsigned char xlat[256];
	pibi_IMAGE my_surf;

	//---------------------------------------------------------
	//reload original every time we update
	lp_host->load();

	//---------------------------------------------------------
	//calc xlat tables
	for(i=0;i<=128;i++)
	{
		xlat[i]=(BYTE)(128.0*pow((i/128.0),1.0+(power/100.0)));
	}
	for(i=129;i<256;i++)
	{
		xlat[i]=(BYTE)(255-xlat[255-i]);
	}

	//---------------------------------------------------------
	//lock bits for access
	lp_host->lock();

	lp_host->pi_BasicImage->virtualize(&my_surf,
		lp_host->width,lp_host->height,
		lp_host->pitch,lp_host->lp_pix);

	//---------------------------------------------------------
	//execute
	lp_host->pi_BasicImage->translate(
		&my_surf,0,0,my_surf.width-1,my_surf.height-1,
		xlat,xlat,xlat,0,0,0);

	//---------------------------------------------------------
	//unlock bits after access
	lp_host->unlock();

	//---------------------------------------------------------
	//force host to refresh its display to reflect changes
	lp_host->refresh();
	//---------------------------------------------------------
	return 1;
}
//=============================================================
LRESULT CALLBACK msg_dlg_effect(
	HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar)
{
	SCROLLINFO si;
	//---------------------------------------------------------
	switch (msg)
	{
	//---------------------------------------------------------
	case WM_INITDIALOG:
		SetWindowText(hDlg,PLUGIN_NAME);
		SetWindowLong(hDlg,GWL_EXSTYLE,WS_EX_TOPMOST|
			GetWindowLong(hDlg,GWL_EXSTYLE));
		memset(&si,0,sizeof(si));
		si.cbSize=sizeof(si);
		si.fMask=SIF_PAGE|SIF_POS|SIF_RANGE;
		si.nPage=100;
		si.nMin=0;
		si.nMax=1000;
		si.nPos=tmp_settings.power;
		SetScrollInfo(GetDlgItem(hDlg,IDC_SCROLL1),SB_CTL,&si,
			TRUE);
		//-----------------------------------------------------
		effect_Execute(tmp_settings.power);
		return 1|lpar;
	//---------------------------------------------------------
	case WM_HSCROLL:
		memset(&si,0,sizeof(si));
		si.cbSize=sizeof(si);
		si.fMask=SIF_PAGE|SIF_POS|SIF_RANGE|SIF_TRACKPOS;
		GetScrollInfo(GetDlgItem(hDlg,IDC_SCROLL1),SB_CTL,&si);
		si.nPage=100;
		si.nMin=0;
		si.nMax=1000;
		//-----------------------------------------------------
		switch(LOWORD(wpar))
		{
		case SB_LEFT:
			si.nPos=si.nMin;
            break;
		case SB_RIGHT:
			si.nPos=si.nMax;
			break;
		case SB_LINELEFT:
			si.nPos-=1;
			break;
		case SB_LINERIGHT:
			si.nPos+=1;
            break;
		case SB_PAGELEFT:
			si.nPos-=100;
            break;
		case SB_PAGERIGHT:
			si.nPos+=100;
			break;
		case SB_THUMBTRACK:
			si.nPos=si.nTrackPos;
			break;
		case SB_ENDSCROLL:
			break;
		}
		//-----------------------------------------------------
		if(si.nPos<si.nMin) si.nPos=si.nMin;
		si.fMask=SIF_PAGE|SIF_POS|SIF_RANGE;
		SetScrollInfo(GetDlgItem(hDlg,IDC_SCROLL1),SB_CTL,&si,
			TRUE);
		GetScrollInfo(GetDlgItem(hDlg,IDC_SCROLL1),SB_CTL,&si);
		tmp_settings.power=si.nPos;
		//-----------------------------------------------------
		effect_Execute(tmp_settings.power);
		break;
	//---------------------------------------------------------
	case WM_COMMAND:
		switch(LOWORD(wpar))
		{
		//-----------------------------------------------------
		case IDOK    :
			//-------------------------------------------------
			effect_Execute(tmp_settings.power);
			EndDialog(hDlg,1);
			break;
		//-----------------------------------------------------
		case IDCANCEL:
			EndDialog(hDlg,0);
			break;
		}
		return 1;
	}
	//---------------------------------------------------------
	return 0;
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////